
// The D/Objective-C Bridge
// Copyright © 2007  Michel Fortin
// <http://michelf.com/d-objc-bridge/>
//
// This program is free software: you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free 
// Software Foundation, either version 3 of the License, or (at your option) 
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT 
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program. If not, see <http://www.gnu.org/licenses/>.

module cocoa.appkit.application;

import cocoa.appkit.window;
import cocoa.appkit.menu;
import cocoa.appkit.responder;
import cocoa.appkit.image;
import cocoa.appkit.menu;

import cocoa.foundation.object;
import cocoa.foundation.string;
import cocoa.foundation.bundle;

import objc.bridge;


class NSApplication : NSResponder {
	mixin ObjcWrap;
	this(objc.runtime.id instance) { super(instance); }
	
	private static NSApplication sharedInstance;

	/**
	 * Create the application object. You can create only one application
	 * per process, and the application may not be destroyed.
	 */
	this() {
		assert(sharedInstance is null, "Cannot create a second NSApplication object.");
		super(invokeObjcSelfClass!(objc.runtime.id, "sharedApplication")());
		sharedInstance = this;
	}
	
	
	/**
	 * Return the shared application instance. This will create a new 
	 * application if it does not yet exist.
	 */
	static NSApplication sharedApplication() {
		if (!sharedInstance) {
			sharedInstance = new NSApplication;
		}
		return sharedInstance;
	}
	
	/**
	 * Run the application event loop.
	 */
	void run() {
		invokeObjcSuper!(void, "run")();
	}
	
	/**
	 * This method is called before starting the event loop.
	 */
	void finishLaunching() {
		invokeObjcSuper!(void, "finishLaunching")();
	}
	
	/**
	 * Activate the appliation.
	 */
	void activate(bool ignoringOtherApps = false) {
		invokeObjcSuper!(void, "activateIgnoringOtherApps:")(ignoringOtherApps);
	}
	
	mixin ObjcBindMethod!(run, void, "run");
	mixin ObjcBindMethod!(finishLaunching, void, "finishLaunching");
	mixin ObjcBindMethod!(activate, void, "activateIgnoringOtherApps:", bool);
	
	
	// Actions

	void terminate(Object sender) {
		invokeObjcSuper!(void, "terminate:")(sender);
	}
	
	void hideOtherApplications(Object sender) {
		invokeObjcSuper!(void, "hideOtherApplications:")(sender);
	}
	
	mixin ObjcBindMethod!(terminate, void, "terminate:", Object);
	mixin ObjcBindMethod!(hideOtherApplications, void, "hideOtherApplications:", Object);
	
	
	// Special Menus

	void mainMenu(NSMenu menu) {
		invokeObjcSuper!(void, "setMainMenu:")(menu);
	}
	
	NSMenu mainMenu() {
		return invokeObjcSuper!(NSMenu, "mainMenu")();
	}
	
	
	// Note: -[NSApplication appleMenu] and -[NSApplication setAppleMenu:] 
	// were removed from headers in Tiger for some reason, but are still 
	// needed to set programatically the application menu, so they are here.

	void appleMenu(NSMenu menu) {
		invokeObjcSuper!(void, "setAppleMenu:")(menu);
	}
	
	NSMenu appleMenu() {
		return invokeObjcSuper!(NSMenu, "appleMenu")();
	}
	

	void windowsMenu(NSMenu menu) {
		invokeObjcSuper!(void, "setWindowsMenu:")(menu);
	}
	
	NSMenu windowsMenu() {
		return invokeObjcSuper!(NSMenu, "windowsMenu")();
	}
	

	void servicesMenu(NSMenu menu) {
		invokeObjcSuper!(void, "setServicesMenu:")(menu);
	}
	
	NSMenu servicesMenu() {
		return invokeObjcSuper!(NSMenu, "servicesMenu")();
	}
	
	
	mixin ObjcBindMethod!(mainMenu, void, "setMainMenu:", NSMenu);
	mixin ObjcBindMethod!(mainMenu, NSMenu, "mainMenu");
	mixin ObjcBindMethod!(appleMenu, void, "setAppleMenu:", NSMenu);
	mixin ObjcBindMethod!(appleMenu, NSMenu, "appleMenu");
	mixin ObjcBindMethod!(windowsMenu, void, "setWindowsMenu:", NSMenu);
	mixin ObjcBindMethod!(windowsMenu, NSMenu, "windowsMenu");
	mixin ObjcBindMethod!(servicesMenu, void, "setServicesMenu:", NSMenu);
	mixin ObjcBindMethod!(servicesMenu, NSMenu, "servicesMenu");


	// Other propreties.
	
	void applicationIconImage(NSImage image) {
		invokeObjcSuper!(void, "setApplicationIconImage:")(image);
	}
	
	NSImage applicationIconImage() {
		return invokeObjcSuper!(NSImage, "applicationIconImage")();
	}
	
	mixin ObjcBindMethod!(applicationIconImage, void, "setApplicationIconImage:", NSImage);
	mixin ObjcBindMethod!(applicationIconImage, NSImage, "applicationIconImage");
	
	
	// Application delegate.
	
	// deleg proprety (in replacement for "delegate" which is a keyword in D).

	void deleg(Object object) {
		invokeObjcSuper!(void, "setDelegate:")(object);
	}

	Object deleg() {
		return invokeObjcSuper!(Object, "delegate")();
	}
	
	mixin ObjcBindMethod!(deleg, void, "setDelegate:", Object);
	mixin ObjcBindMethod!(deleg, Object, "delegate");


	// NSBundle AppKit Additions
	
	static bool loadNib(NSString name, Object owner) {
		return NSBundle.invokeObjcSuperClass!(bool, "loadNibNamed:owner:")(name, owner);
	}
	
	static bool loadNib(char[] name, Object owner) {
		return loadNib(name.toNSString(), owner);
	}
	
	
	enum TerminateReply { CANCEL = 0, NOW = 1, LATER = 2 }

}



void NSApplicationMain(char[][] args) {
	auto app = new NSApplication;
	NSApplication.loadNib("myMain", app);
	app.run;
}
