
// The D/Objective-C Bridge
// Copyright © 2007  Michel Fortin
// <http://michelf.com/d-objc-bridge/>
//
// This program is free software: you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free 
// Software Foundation, either version 3 of the License, or (at your option) 
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT 
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program. If not, see <http://www.gnu.org/licenses/>.

module cocoa.foundation.array;

import cocoa.foundation.object;
import cocoa.foundation.string;
import cocoa.foundation.range;
import cocoa.foundation.enumerator;

import objc.bridge;
import objc.runtime;


/**
 * Create an NSArray from a standard D array.
 */
NSArray toNSArray(Object[] objects) {
	return new NSArray(objects);
}


/**
 * Wrapper for Objective-C class NSArray.
 *
 * Subclassing notes: NSArray is a cluster class. If you subclass it, 
 * you need to create your own Objective-C instance first with objcAlloc and 
 * pass it as the argument for the NSArray constructor. Otherwise, if 
 * you call another NSArray constructor which creates the Objective-C 
 * instance for you, that instance will not be a capsule for your subclass.
 */
class NSArray : NSObject {
	mixin ObjcClusterWrap;
	this(id instance) { super(instance); }
	
		
	/** Gives the lenght of the array. [self count] */
	uint length() {
		return invokeObjcSuper!(uint, "count")();
	}
	
	/** Gives the object for a specified index. [self objectAtIndex:index] */
	Object opIndex(uint index) {
		return invokeObjcSuper!(Object, "objectAtIndex:")(index);
	}
	
	mixin ObjcBindMethod!(length, uint, "count");
	mixin ObjcBindMethod!(opIndex, Object, "objectAtIndex:", uint);
	
	
	// NSExtendedArray
	
	/** Gives an array with object is appended at the end of this one. */
	NSArray arrayByAddingObject(Object object) {
		return invokeObjcSuper!(NSArray, "arrayByAddingObject:")(object);
	}
	
	/** Gives an array where objects are appended at the end of this one. */
	NSArray opCat(NSArray objects) {
		return invokeObjcSuper!(NSArray, "arrayByAddingObjectsFromArray:")(objects);
	}
	
	/** ditto */
	final NSArray opCat(Object[] objects) {
		if (objects.length == 1)  return arrayByAddingObject(objects[0]);
		return opCat(objects.toNSArray());
	}
	
	/** Tell if a value is found in the array (using equality operator). */
	bool opIn_r(Object object) {
		return invokeObjcSuper!(bool, "containsObject:")(object);
	}
	
	/** Gives the index for a given object (using equality operator). */
	uint indexOf(Object object) {
		return invokeObjcSuper!(bool, "indexOfObject:")(object);
	}
	
	/** Tell if this array is equal to other array. */
	bool isEqualToArray(NSArray other) {
		return invokeObjcSuper!(bool, "isEqualToArray:")(other);
	}
	
	uint opEquals(Object[] objects) {
		if (this.length != objects.length)  return false;
		if (length == 0)                    return true;
		return isEqualToArray(objects.toNSArray());
	}
	
	/** Gives a slice of the complete array, this effectivly returns the array. */
	NSArray opSlice() {
		return this;
	}
	
	/** Slice array from an index to another index. */
	NSArray opSlice(uint from, uint to) {
		return invokeObjcSuper!(NSArray, "subarrayWithRange:")(NSRange[from..to]);
	}
	
	alias NSObject.opEquals opEquals;
	
	mixin ObjcBindMethod!(arrayByAddingObject, NSArray, "arrayByAddingObject:", Object);
	mixin ObjcBindMethod!(opCat, NSArray, "arrayByAddingObjectsFromArray:", NSArray);
	mixin ObjcBindMethod!(opIn_r, bool, "containsObject:", Object);
	mixin ObjcBindMethod!(indexOf, uint, "indexOfObject:", Object);
	mixin ObjcBindMethod!(isEqualToArray, bool, "isEqualToArray:", NSArray);
	
	// Special forwarder function for opSlice converting NSRange 
	// in a from..to form.
	private static id objcForward_subarrayWithRange_(id self, SEL sel, NSRange range) {
		return objc.exceptionbridge.callD( 
			objcWrapInstance(self).opSlice(range.location, range.end).objcInstance
			);
	}
	mixin ObjcMethodInfo!(objcForward_subarrayWithRange_, NSArray, "subarrayWithRange:", NSRange);
	
	/** Write array to path as a property list. */
	bool writeToFile(NSString path, bool atomically) {
		return invokeObjcSuper!(bool, "writeToFile:atomically:")(path, atomically);
	}
	
//	/** Write array to URL as a property list. */
//	bool writeToURL(NSURL url, bool atomically) {
//		return invokeObjcSuper!(bool, "writeToURL:atomically:")(url, atomically);
//	}
	
	mixin ObjcBindMethod!(writeToFile, bool, "writeToFile:atomically:", NSString, bool);
//	mixin ObjcBindMethod!(writeToURL, bool, "writeToURL:atomically:", NSURL, bool);


	// (NSArrayCreation)

	/** Create an empty array. */
	this() {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "init")();
	}
	
	/** Create an array filled with all the objects from other array. */
	this(NSArray array) {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "initWithArray:")(array);
	}
	
	/** Create an array by reading property list file at path. */
	this(NSString path) {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "initWithContentOfFile:")(path);
	}

//	/** Create an array by reading property list file at URL. */
//	this(NSURL url) {
//		this(objcClusterAlloc);
//		objcInstance = invokeObjcSuper!(id, "initWithContentOfURL:")(url);
//	}

	/** Create an array with an array of objects. */
	this(Object[] objects) {
		this(objects.toObjcInstances());
	}
	
	/** Create an array with an array of Objective-C object instances. */
	this(id[] objects) {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "initWithObjects:count:")(objects.ptr, objects.length);
	}

	
	// Foreach loop support.
	
	int opApply(int delegate(ref Object) dg) {
		uint l = length;
		for (uint i = 0; i < l; i++) {
			Object object = this[i];
			int broken = dg(object);
			if (broken) return broken;
		}
		return 0;
	}
	
	int opApply(int delegate(ref uint, ref Object) dg) {
		uint l = length;
		for (uint i = 0; i < length; i++) {
			Object object = this[i];
			uint index = i;
			int broken = dg(index, object);
			if (broken) return broken;
		}
		return 0;
	}
	
	int opApplyReverse(int delegate(ref Object) dg) {
		uint i = length;
		while (i-- < 0) {
			Object object = this[i];
			int broken = dg(object);
			if (broken) return broken;
		}
		return 0;
	}
	
	int opApplyReverse(int delegate(ref uint, ref Object) dg) {
		uint i = length;
		while (i-- < 0) {
			Object object = this[i];
			uint index = i;
			int broken = dg(index, object);
			if (broken) return broken;
		}
		return 0;
	}
}


/**
 * Wrapper for Objective-C class NSMutableArray.
 *
 * Subclassing notes: NSMutableArray is a cluster class. If you subclass it, 
 * you need to create your own Objective-C instance first with objcAlloc and 
 * pass it as the argument for the NSMutableArray constructor. Otherwise, if 
 * you call another NSMutableArray constructor which creates the Objective-C 
 * instance for you, that instance will not be a capsule for your subclass.
 */
class NSMutableArray : NSArray {
	mixin ObjcClusterWrap;
	this(id instance) { super(instance); }
	
	/** Append object at the end of array. */
	void add(Object object) {
		invokeObjcSuper!(void, "addObject:")(object);
	}
	
	/** Insert object at index in array. */
	void insert(Object object, uint index) {
		invokeObjcSuper!(void, "insertObject:atIndex:")(object, index);
	}
	
	/** Remove last object in array. */
	void removeLast() {
		invokeObjcSuper!(void, "removeLastObject")();
	}
	
	/** Remove object at index. */
	void remove(uint index) {
		invokeObjcSuper!(void, "removeObjectAtIndex:")(index);
	}
	
	/** Set object at index to replacement. */
	void opIndexAssign(uint index, Object replacement) {
		invokeObjcSuper!(void, "replaceObjectAtIndex:withObject:")(index, replacement);
	}
	
	mixin ObjcBindMethod!(add, void, "addObject:", Object);
	mixin ObjcBindMethod!(insert, void, "insertObject:atIndex:", Object, uint);
	mixin ObjcBindMethod!(removeLast, void, "removeLast");
	mixin ObjcBindMethod!(remove, void, "removeObjectAtIndex:", uint);
	mixin ObjcBindMethod!(opIndexAssign, void, "replaceObjectAtIndex:withObject:", uint, Object);
	
	
	// (NSExtendedMutableArray)
	
	void opCatAssign(NSArray objects) {
		invokeObjcSuper!(void, "addObjectsFromArray:")(objects);
	}
	
	void opCatAssign(Object[] objects) {
		if (objects.length == 0)  return;
		if (objects.length == 1)  add(objects[0]);
		else                      opCatAssign(objects.toNSArray());
	}
	
	void swap(uint index1, uint index2) {
		invokeObjcSuper!(void, "exchangeObjectAtIndex:withObjectAtIndex:")(index1, index2);
	}
	
	void removeAll() {
		invokeObjcSuper!(void, "removeAllObjects")();
	}

	void remove(Object object) {
		invokeObjcSuper!(void, "removeObject:")(object);
	}

	void removeEqual(Object object) {
		invokeObjcSuper!(void, "removeObjectIdenticalTo:")(object);
	}

	mixin ObjcBindMethod!(opCatAssign, void, "addObjectsFromArray:", NSArray);
	mixin ObjcBindMethod!(swap, void, "exchangeObjectAtIndex:withObjectAtIndex:", uint, uint);
	mixin ObjcBindMethod!(removeAll, void, "removeAllObjects");
	mixin ObjcBindMethod!(remove, void, "removeObject:", Object);
	mixin ObjcBindMethod!(removeEqual, void, "removeObjectIdenticalTo:", Object);
	

	void removeAll(NSArray objects) {
		invokeObjcSuper!(void, "removeObjectsInArray:")(objects);
	}
	
	void removeAll(NSRange range) {
		invokeObjcSuper!(void, "removeObjectsInRange:")(range);
	}
		
	mixin ObjcBindMethod!(removeAll, void, "removeObjectsInArray:", NSArray);
	mixin ObjcBindMethod!(removeAll, void, "removeObjectsInRange:", NSRange);
	
	
	void opSliceAssign(NSArray array, uint from, uint to) {
		invokeObjcSuper!(void, "replaceObjectsInRange:withObjectsFromArray:")(NSRange[from..to], array);
	}
		
	// Special forwarder function for opSliceAssign converting NSRange 
	// in a from..to form and adjusting argument order.
	private static void objcForward_replaceObjectsInRange_withObjectsFromArray_(id self, SEL sel, NSRange range, id array) {
		return objc.exceptionbridge.callD( 
			objcWrapInstance(self).opSliceAssign(NSArray.objcWrapInstance(array), range.location, range.end)
			);
	}
	mixin ObjcMethodInfo!(objcForward_replaceObjectsInRange_withObjectsFromArray_, void, "subarrayWithRange:", NSRange, NSArray);


	void opAssign(NSArray array) {
		invokeObjcSuper!(void, "setArray:")(array);
	}
	
	void sort(int function(id, id, void*) compare, void* context) {
		invokeObjcSuper!(void, "sortUsingFunction:context:")(compare, context);
	}
	
	final void sort(int delegate(Object, Object) compare) {
		static int comparator(id a, id b, void* context) {
			auto compare = *cast(int delegate(Object, Object)*)context;
			Object a1 = objc.type.decapsulate!(Object)(a);
			Object a2 = objc.type.decapsulate!(Object)(b);
			return compare(a1, a2);
		}
		sort(&comparator, cast(void*)&compare);
	}
	
	final void sort() {
		static int comparator(id a, id b, void*) {
			Object a1 = objc.type.decapsulate!(Object)(a);
			Object a2 = objc.type.decapsulate!(Object)(b);
			return a1.opCmp(a2);
		}
		sort(&comparator, null);
	}
		
	mixin ObjcBindMethod!(opAssign, void, "setArray:", NSArray);
	mixin ObjcBindMethod!(sort, void, "sortUsingFunction:context:", int function(id, id, void*), void*);
	

	// NSMutableArrayCreation
	
	this(uint capacity) {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "initWithCapacity:")(capacity);
	}

	
	// Repeating NSArrayCreation

	/** Create an empty array. */
	this() {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "init")();
	}
	
	/** Create an array filled with all the objects from other array. */
	this(NSArray array) {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "initWithArray:")(array);
	}
	
	/** Create an array by reading property list file at path. */
	this(NSString path) {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "initWithContentOfFile:")(path);
	}

//	/** Create an array by reading property list file at URL. */
//	this(NSURL url) {
//		this(objcClusterAlloc);
//		objcInstance = invokeObjcSuper!(id, "initWithContentOfURL:")(url);
//	}

	/** Create an array with an array of objects. */
	this(Object[] objects) {
		this(objects.toObjcInstances());
	}
	
	/** Create an array with an array of Objective-C object instances. */
	this(id[] objects) {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "initWithObjects:count:")(objects.ptr, objects.length);
	}

}


// Convert array of objects to an array of Objective-C object instances.
package id[] toObjcInstances(object.Object[] objects) {
	id[] instances;
	instances.length = objects.length;
	foreach (int i, Object object; objects) {
		instances[i] = objc.type.encapsulate!(Object)(object);
	}
	return instances;
}

// Convert array of strings to an array of Objective-C object instances.
package id[] toObjcInstances(char[][] strings) {
	id[] instances;
	instances.length = strings.length;
	foreach (int i, char[] string; strings) {
		instances[i] = string.toNSString().objcInstance;
	}
	return instances;
}


import cocoa.foundation.autoreleasepool;

unittest {
	scope NSAutoreleasePool pool = new NSAutoreleasePool;

	auto s1 = "1".toNSString();
	auto s2 = "2".toNSString();
	auto o1 = new Object;
	auto array1 = toNSArray([s1, s2]);
	auto array2 = toNSArray([s2, s1]);
	auto array3 = array1 ~ array2 ~ new NSArray([o1]);
	
	assert(array3[0] is s1);
	assert(array3[1] is s2);
	assert(array3[2] is s2);
	assert(array3[3] is s1);
	assert(array3[4] is o1);
	
	assert(array1 != array2);
	assert(array1 ~ [new Object]);
	
	foreach (i, o; array1) {
		// Nothing to do really.
	}
	
	try {
		auto result = array3[5];
		assert(0);
	} catch {
		// Caught out-of-bound exception.
	}

}

unittest {
	scope NSAutoreleasePool pool = new NSAutoreleasePool;
	NSMutableArray array = new NSMutableArray;
	
	array ~= ["obj1".toNSString()];
	array ~= ["obj2".toNSString()];

	assert(array.length == 2);
	assert("obj2".toNSString() in array);
	assert(array[].length == 2);
	assert(array[0..1] == ["obj1".toNSString()]);
	
	array ~= array;
	assert(array.length == 4);

}