
// The D/Objective-C Bridge
// Copyright © 2007  Michel Fortin
// <http://michelf.com/d-objc-bridge/>
//
// This program is free software: you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free 
// Software Foundation, either version 3 of the License, or (at your option) 
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT 
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program. If not, see <http://www.gnu.org/licenses/>.

module cocoa.foundation.exception;

import cocoa.foundation.object;
import cocoa.foundation.string;
import cocoa.foundation.dictionary;

import objc.bridge;
import objc.exceptionbridge;


class NSException : NSObject {
	mixin ObjcWrap;
	this(objc.runtime.id instance) { super(instance); }
	
	this(NSString name, NSString reason, NSDictionary userInfo) {
		this(objcAlloc);
		objcInstance = invokeObjcSuper!(objc.runtime.id, "initWithName:reason:userInfo:")(name, reason, userInfo);
	}
	
	this(char[] name, char[] reason, NSDictionary userInfo) {
		this(name.toNSString(), reason.toNSString(), userInfo);
	}
	
	
	NSString name() {
		return invokeObjcSuper!(NSString, "name")();
	}
	mixin ObjcBindMethod!(name, NSString, "name");
	
	
	NSString reason() {
		return invokeObjcSuper!(NSString, "reason")();
	}
	mixin ObjcBindMethod!(reason, NSString, "reason");
	
	
	NSDictionary userInfo() {
		return invokeObjcSuper!(NSDictionary, "userInfo")();
	}
	mixin ObjcBindMethod!(userInfo, NSDictionary, "userInfo");
	

	// -- Exception Bridge Support
	
	/**
	 * Raise an exception in the Objective-C realm.
	 *
	 * Note: You should throw exceptions with the throw keyword when in D code 
	 * instead of calling raise. This method is private and is only for use 
	 * by the D/ObjC exception bridge.
	 */
	private void raise() {
		// Bypass invokeObjcSuper template to avoid exception being rewrapped
		// and rethrown as a D object.
		objc.msg.sendSuper!(void)(&context, objc.method.selector!("raise"));
	}
	
	/**
	 * Return the throwable object for this exception. If the exception is an
	 * enclosure for a D object, this method will return the enclosed object,
	 * otherwise it'll return itself as this implementation does.
	 */
	protected Object throwable() {
		return this;
	}
	
	/**
	 * Enclose object within an NSException.
	 */
	private static NSException enclose(Object object) {
		return new D_ExceptionEnclosure(object);
	}
	
	/**
	 * Throw an ObjC NSException as a D object. 
	 * Callback for the D/ObjC exception bridge.
	 */
	private static void throwException(objc.runtime.id exception) {
		throw objcWrapInstance(exception).throwable;
	}

	/**
	 * Raise object as an ObjC exception. 
	 * Callback for the D/ObjC exception bridge.
	 */
	private static void raiseException(Object o) {
		auto e = cast(NSException)o;
		if (!e)  e = enclose(o);
		e.raise;
	}

}


extern (C) {
	// Generic Exception Names (NSString)

	extern objc.runtime.id NSGenericException;
	extern objc.runtime.id NSRangeException;
	extern objc.runtime.id NSInvalidArgumentException;
	extern objc.runtime.id NSInternalInconsistencyException;
	extern objc.runtime.id NSMallocException;
	extern objc.runtime.id NSObjectInaccessibleException;
	extern objc.runtime.id NSObjectNotAvailableException;
	extern objc.runtime.id NSDestinationInvalidException;
	extern objc.runtime.id NSPortTimeoutException;
	extern objc.runtime.id NSInvalidSendPortException;
	extern objc.runtime.id NSInvalidReceivePortException;
	extern objc.runtime.id NSPortSendException;
	extern objc.runtime.id NSPortReceiveException;
	extern objc.runtime.id NSOldStyleException;
}


/**
 * Special exception class used to enclose a thrown D object into an 
 * NSException while passing through Objective-C code.
 */
private class D_ExceptionEnclosure : NSException {

	private Object object;
	
	this(Object object) {
		this.object = object;
		auto name = "D(" ~ object.classinfo.name ~ ")";
		
		// Set message if enclosure is of class Exception.
		auto exception = cast(Exception)object;
		auto desc = exception ? exception.msg : "";
		
		super(name, desc, null);
	}
	
	override Object throwable() {
		return object;
	}
	
}


// Initializing the exception bridge with our exception handlers.
static this() {
	objc.exceptionbridge.throwExceptionFunc = &NSException.throwException;
	objc.exceptionbridge.raiseExceptionFunc = &NSException.raiseException;
}


import cocoa.foundation.autoreleasepool;

unittest {
	scope NSAutoreleasePool pool = new NSAutoreleasePool;

	void testThrow() {
		throw new Exception("D_ExceptionEnclosureTest");
	}
	
	try {
		callObjc( callD( testThrow() ) );
		assert(0);
	} catch (Exception e) {
		// successfully caught D exception.
	} catch {
		assert(0);
	}
	
	void testRaise() {
		auto e = new NSException("NSExceptionTest", "Test", null);
		e.raise;
	}

	try {
		callObjc( callD( testRaise() ) );
		assert(0);
	} catch (NSException e) {
		// successfully caught Objective-C exception.
	} catch {
		assert(0);
	}

}
