
// The D/Objective-C Bridge
// Copyright © 2007  Michel Fortin
// <http://michelf.com/d-objc-bridge/>
//
// This program is free software: you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free 
// Software Foundation, either version 3 of the License, or (at your option) 
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT 
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program. If not, see <http://www.gnu.org/licenses/>.

/**
 * This module enable passing any D object as an Objective-C object.
 *
 * Encapsulated objects appear in Objective-C as of class D_Object, a direct 
 * descendent of NSObject which reimplements basic methods so they are 
 * forwarded to the actual D object. These three methods are forwarded:
 *
 *  - hash         (mapped to Object.toHash)
 *  - description  (mapped to Object.toString)
 *  - isEqual:     (mapped to Object.opEquals)
 *
 * Retaining an encapsulated objects on the Objective-C side will prevent
 * prevent the D garbage collector from deleting objects prematuarlly.
 */
module objc.capsule;

import objc.type;
import objc.typecoding;
import objc.runtime;
import objc.method;
import objc.methodlist;
import objc.subclassinit;
import objc.exceptionbridge;
import objc.wrapper;

import std.gc;


// -- Callbacks

id function(char[] string) encapsulateStringFunc = &encapsulateString;


// -- Object wrapper for non-capsuled types.

id encapsulateObject(Object object) {
	if (object is null) {
		return null;
	}
	if (auto objcObject = cast(ObjcObject)object) {
		return objcObject.objcInstance;
	}
	return Capsule.encapsulate(object);
}


Object decapsulateObject(id object) {
	if (object is null) {
		return null;
	}
	if (!isCapsule(object)) {
		return objc.wrapper.wrapInstance(object);
	}
	return Capsule.decapsulate(object);
}

static this() {
	encapsulateObjectFunc = &encapsulateObject;
	decapsulateObjectFunc = &decapsulateObject;
}


private objc_class capsuleClassDef; // Objective-C class D_Object
private objc_variable_method_list!(4) instanceMethodListDef;

private objc_class capsuleMetaClassDef;
private objc_variable_method_list!(0) classMethodListDef;


/**
 * Objective-C class structure for a capsule.
 */
Class capsuleClass() {
	if (capsuleClassDef.isa == null) {
		initCapsule;
	}
	return &capsuleClassDef;
}


/**
 * Check if object is a capsule.
 */
bool isCapsule(id object)
in {
	assert(object !is null);
}
body {
	Class objcClass = object.isa;
	
	do {
		if (objcClass == &capsuleClassDef)  return true;
		objcClass = objcClass.super_class;
	} while (objcClass != null);
	
	return false;
}



private void initCapsule() {
	with (instanceMethodListDef.method_list[0]) {
		method_name  = selector!("hash");
		method_types = methodTypeEncode!(uint, id, SEL).ptr;
		method_imp   = cast(objc.runtime.IMP)&Capsule.hash;
	}
	with (instanceMethodListDef.method_list[1]) {
		method_name  = selector!("description");
		method_types = methodTypeEncode!(id, id, SEL).ptr;
		method_imp   = cast(objc.runtime.IMP)&Capsule.description;
	}
	with (instanceMethodListDef.method_list[2]) {
		method_name  = selector!("isEqual");
		method_types = methodTypeEncode!(bool, id, SEL, id).ptr;
		method_imp   = cast(objc.runtime.IMP)&Capsule.isEqual;
	}
	with (instanceMethodListDef.method_list[3]) {
		method_name  = selector!("dealloc");
		method_types = methodTypeEncode!(void, id, SEL).ptr;
		method_imp   = cast(objc.runtime.IMP)&Capsule.dealloc;
	}

	const objc_method_list*[1] instanceMethodLists = [cast(objc_method_list*)&instanceMethodListDef];
	const objc_method_list*[1] classMethodLists = [cast(objc_method_list*)&classMethodListDef];
	
	capsuleClassDef.instance_size = Capsule.sizeof;

	subclassInit("D_Object".ptr, objc_lookUpClass("NSObject".ptr), 
		&capsuleClassDef, &capsuleMetaClassDef, 
		instanceMethodLists, classMethodLists);
}


id createCapsule(Object object, Class capsuleClass) {
	return Capsule.encapsulate(object, capsuleClass);
}


Object removeCapsule(id capsule) {
	return Capsule.decapsulate(capsule);
}


private struct Capsule {
	Class isa;
	Object object;
	
	static id encapsulate(Object object) {
		return encapsulate(object, capsuleClass);
	}
	
	static id encapsulate(Object object, Class capsuleClass) 
	in {
		assert(object !is null);
	}
	body {
		auto capsule = cast(Capsule*)class_createInstance(capsuleClass, 0);
		
		capsule.object = object;
		addRoot(&capsule.object);
		
		return capsule.self;
	}
	
	static Object decapsulate(id object) 
	in {
		assert(object !is null);
		assert(isCapsule(object));
	}
	body {
		auto capsule = cast(Capsule*)object;
		return capsule.object;
	}
	
	id self() {
		return cast(id)this;
	}
	
	uint hash(SEL sel) {
		return callD( object.toHash() );
	}
	
	id description(SEL sel) {
		return callD( encapsulateStringFunc(object.toString()) );
	}
	
	uint isEqual(SEL sel, id other) {
		if (isCapsule(other)) {
			// Just call [super isEqual];
			objc_super context = superContext;
			objc.msg.sendSuper!(void)(&context, sel);
		} else {
			return callD( object.opEquals(decapsulate(other)) );
		}
	}
	
	void dealloc(SEL sel) {
		removeRoot(&object);
		
		objc_super context = superContext;
		objc.msg.sendSuper!(void)(&context, sel);
	}
	
	objc_super superContext() {
		objc_super context;
		context.receiver = self;
		context.super_class = capsuleClassDef.super_class;
		return context;
	}
	
}


/**
 * Default string encapsulator implementation. Simply copy the string 
 * into an autoreleased NSString.
 */
id encapsulateString(char[] string) {
	static id stringClass;
	SEL selector = objc.method.selector!("initWithBytes:length:encoding:");
	const uint NSUTF8StringEncoding = 4;

	if (!stringClass) {
		stringClass = cast(id)objc_lookUpClass("NSString".ptr);
	}
	return objc.msg.send!(id)(stringClass, selector, string.ptr, string.length, NSUTF8StringEncoding);
}
