
// The D/Objective-C Bridge
// Copyright © 2007  Michel Fortin
// <http://michelf.com/d-objc-bridge/>
//
// This program is free software: you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free 
// Software Foundation, either version 3 of the License, or (at your option) 
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT 
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program. If not, see <http://www.gnu.org/licenses/>.

/**
 * Allow declaration and gathering of instance methods, class methods and 
 * initializer methods for a given D class.
 */
module objc.methodlist;

import objc.runtime;
import objc.method;

import std.traits;
import std.typetuple;


enum MethodKind {
	INSTANCE_INIT,
	INSTANCE,
	CLASS,
}


struct objc_variable_method_list(int count) {

	/// Reserved for future use
	void* obsolete;

	/// An integer specifying the number of methods in the method list 
	/// array
	int method_count = count;

	/// An array of objc_method data structures
	objc.runtime.objc_method[count] method_list;
}


// Make sure objc_variable_method_list grows when count increase. This will
// detects an old bug with GDC and static arrays within structs.
static assert(
	(objc_variable_method_list!(2)).sizeof + objc_method.sizeof == 
	(objc_variable_method_list!(3)).sizeof);


/**
 * Adding a MemberObjcMethod struct to a class will make it retrievable by
 * the templates below.
 */
struct MemberObjcMethod(char[] name, char[] type, alias imp, MethodKind kind) {
	const static MethodKind methodKind = kind;
	const static objc.runtime.objc_method methodInfo = {
		method_name:  cast(objc.runtime.SEL)name.ptr, // registration needed
		method_types: type.ptr,
		method_imp:   cast(objc.runtime.IMP)&imp
	};
}


/**
 * Tell if a type is a method info struct of a given kind.
 */
template objcIsMethodInfoOfKind(T, MethodKind kind) {
	static if (
		is(typeof(T.methodInfo) == objc.runtime.objc_method) &&
		is(typeof(T.methodKind) == MethodKind) &&
		T.methodKind == kind
		)
	{ 
		const bool objcIsMethodInfoOfKind = true;
	} else {
		const bool objcIsMethodInfoOfKind = false;
	}
}


/**
 * Template to aggregate the methods in a tuple inside an array.
 */
template objcMethodsOf(T, MethodKind kind) {
	static const objc.runtime.objc_method[] objcMethodsOf = 
		objcMethodsInTuple!(kind, FieldTypeTuple!(T));
}

template objcMethodsInTuple(MethodKind kind, T...) {
	static if (T.length == 0) {
		static const objc.runtime.objc_method[] objcMethodsInTuple = [];
	}
	else static if (objcIsMethodInfoOfKind!(T[0], kind)) {
		static const objc.runtime.objc_method[] objcMethodsInTuple = 
			T[0].methodInfo ~ objcMethodsInTuple!(kind, T[1..$]);
	}
	else {
		static const objc.runtime.objc_method[] objcMethodsInTuple =
			objcMethodsInTuple!(kind, T[1..$]);
	}
}


/**
 * Template to count the number of methods in a tuple.
 */
template objcCountMethodsOf(T, MethodKind kind) {
	static const int objcCountMethodsOf = 
		objcCountMethodsInTuple!(kind, FieldTypeTuple!(T));
}

/** ditto */
template objcCountMethodsInTuple(MethodKind kind, T...) {
	static if (T.length == 0) {
		static const int objcCountMethodsInTuple = 0;
	}
	else static if (objcIsMethodInfoOfKind!(T[0], kind)) {
		static const int objcCountMethodsInTuple = 1 + objcCountMethodsInTuple!(kind, T[1..$]);
	}
	else {
		static const int objcCountMethodsInTuple = objcCountMethodsInTuple!(kind, T[1..$]);
	}
}


/**
 * Gives a pointer to an objc_method_list structure containing the list of 
 * methods for T.
 */
objc.runtime.objc_method_list* objcMethodListOf(T, MethodKind kind)() {
	const int count = objcCountMethodsOf!(T, kind);
	
	static if (count > 0) {
		static objc_variable_method_list!(count) methods;
		methods.method_list.ptr[0..count] = objcMethodsOf!(T, kind)[0..count];
		
		for (int i = 0; i < methods.method_count; i++) {
			with (methods.method_list.ptr[i]) {
				method_name = objc.runtime.sel_registerName(method_name);
			}
		}
		
		return cast(objc.runtime.objc_method_list*)&methods;
	} else {
		return null;
	}
}





// For unittest
private {

	void* ptr() { return null; }

	class X {
		int z;
		
		static X objcWrap(objc.runtime.id) {
			return new X;
		}
				
		template ObjcAddMethod(char[] method, R, char[] sel, A...) {
			mixin("
			R "~ method ~"(A args) {
				return objc.method.msgSend!(R, A)(
					null, objc.method.selector!(sel), args);
			}
			private final R msgForwardCall(A args) {
				return "~ method ~"(args);
			}
			alias objc.method.msgReceive!(R, objcWrap, msgForwardCall, A) msgReceive;
			objc.methodlist.MemberObjcMethod!(sel, objc.type.methodTypeEncode!(R, objc.runtime.id, objc.runtime.SEL, A), msgReceive, MethodKind.INSTANCE) a;
			");
		}
		
		mixin ObjcAddMethod!("method", int, "method");
		mixin ObjcAddMethod!("method", int, "method:", int);
	}
	
	class Y : X {
	
		override int method() { return 1; }
		override int method(int i) { return i; }
	
	}
	
	class Z : X {
	
		mixin ObjcAddMethod!("method", int, "method");
	
	}

}

unittest {
	// X
	
	// Note: Unsing method_list.ptr to avoid bound checking; GDC has a bug with 
	// static arrays in struct.

	const int count = objcCountMethodsOf!(X, MethodKind.INSTANCE);
	objc_variable_method_list!(count) methods;
	static assert(count == 2);
	
	methods.method_list.ptr[0..count] = objcMethodsOf!(X, MethodKind.INSTANCE)[0..count];
	
	for (int i = 0; i < methods.method_count; i++) {
		with (methods.method_list.ptr[i]) {
			method_name = objc.runtime.sel_registerName(method_name);
//			printf("%s  %s  %d\n", method_name, method_types, method_imp);
		}
	}
	
	objc.runtime.objc_method_list* list = objcMethodListOf!(X, MethodKind.INSTANCE);
	assert(list.method_count == 2);
	

	// Y

	static assert(objcCountMethodsOf!(Y, MethodKind.INSTANCE) == 0);
	assert(objcMethodListOf!(Y, MethodKind.INSTANCE) == null);
	

	// Z

	static assert(objcCountMethodsOf!(Z, MethodKind.INSTANCE) == 1);
	objc.runtime.objc_method_list* listZ = objcMethodListOf!(Z, MethodKind.INSTANCE);
	assert(listZ.method_count == 1);
	
	static assert(objcCountMethodsOf!(Z, MethodKind.INSTANCE_INIT) == 0);
	static assert(objcCountMethodsOf!(Z, MethodKind.CLASS) == 0);
	
}


